
type ''pred block = ''pred set * ''pred set (* positive atoms, negative atoms. *)

type eqvars =
    (int -m> int set) (* map each integer i to the set of integers j>i
		       that normalize to i. *)
    ;

datatype (''pred, ''fun) clause =
    C_BLOCK of ''pred block
  | C_COMPLEX of (* complex clause *)
    ''pred block (* not really a block; if block is {+P1, ..., +Pm, -Q1, ..., -Qn},
		  this represents +P1(f(xis)),...,+Pm(f(xis)),-Q1(f(xis)),...,-Qn(f(xis));
		  must be non-empty*)
    * ''fun * eqvars (* function heading complex atom, plus rewrites. *)
    * (int * ''pred block) list (* map from normal variable to block on this variable,
				 coded as a list, in increasing order of variables. *)
    (* e.g., P (x), P (y) => P (c (x,y)) is:
     C_COMPLEX (block[+P], c, {}, [(1, block[-P]), (2, block[-P])])
     where block[+P] = ({P}, {}), block[-P] = ({}, {P}).
     *)
    ;

datatype id = ID_NAME of string
  | ID_NUM of int
  | ID_NONE;

type ''pred deduction = id * id * ''pred;

datatype ''pred justif = J_DED of ''pred deduction
  | J_AX of id;

datatype (''pred, ''fun) branch =
    BRANCH of |[ clauses : id -m> (''pred, ''fun) clause,
		 present : id -m> (''pred, ''fun) clause,
		 (* subset of clauses that really are on the branch.
		  Clauses that are not are, first, those that are pending,
		  in one of the queues.  Second, those that are
		  just used for forward subsumption tests.  These
		  are clauses we have already dealt with in branches
		  to the left of this one. *)
		 (* index maps predicates P to the set of clauses on the branch
		  with maximal literal of the form P(...): *)
		 pos_index : ''pred -m> id list * (''fun -m> id list),
		 neg_index : ''pred -m> id list * (''fun -m> id list),
		 (* Each such clause may be of one of two forms:
		  - clauses with maximal literal of the form P(x) (we list their ids).
		    These are all blocks.
		  - clauses with maximal literal of the form P(f(x1,...,xn))
		    (x1,...,xn not necessarily pairwise distinct).  These are
		    never blocks. *)
		 simple_queue : ((''pred, ''fun) clause * ''pred justif) list,
		 complex_queue : ((''pred, ''fun) clause list * ''pred justif) list,
		 deductions : id -m> ''pred deduction
		 ]|;

type (''pred, ''fun) tableau = (''pred, ''fun) branch list;

extern val tableau_true : (''pred, ''fun) tableau;
(* The tableau you have to start with, adding new clauses
 with add_clauses, and doing deduction by expand_tableau,
 until you possibly reach tableau_false (contradiction). *)
extern val tableau_false : (''pred, ''fun) tableau;

exception ContradictionEvt;

extern val expand_tableau : (''_pred * ''_pred -> bool) -> (''_pred, ''_fun) tableau -> (''_pred, ''_fun) tableau;
extern val add_clauses : (''pred, ''fun) tableau *
	   ((''pred, ''fun) clause * ''pred justif) list -> (''pred, ''fun) tableau;
(* clause list should not contain either the empty clause or a tautology. *)
extern val add_clause_lists : (''_pred, ''_fun) tableau *
	   ((''_pred, ''_fun) clause list * ''_pred justif) list -> (''_pred, ''_fun) tableau;
(* list of clause lists may contain the empty clause or a tautology.
 However no clause in a clause list should be empty. *)

type ''fun sigture = ''fun -m> int;

extern val print_clause : string sigture * |[put : string -> unit, ... : 'a]| ->
	   (string, string) clause -> unit;
extern val print_branch : string sigture * |[put : string -> unit, ... : 'a]| ->
	   (string, string) branch -> unit;
extern val print_tableau : string sigture * |[put : string -> unit, ... : 'a]| ->
	   (string, string) tableau -> unit;
