
open "gclauses_h"

fun gclause_of_definite (DCLAUSE (name, env, h, body)) =
    GCLAUSE (name, env, (true, h)::[(false, a) | a in list body]);

fun definite_of_gclause (GCLAUSE (name, env, pregcl)) =
    let val head = ref NONE
	val body = [a
		   | (sign, a) in list pregcl
		       such that (if sign
				      then case !head of
					       NONE => (head := SOME a;
							false)
					     | _ => raise NonDefinite
				  else true)]
    in
	case !head of
	    SOME h => DCLAUSE (name, env, h, body)
	  | _ => raise NonDefinite
    end;

fun avars (ATOM (_, l)) = union {tvars t | t in list l};

fun print_atom (f as |[put, ...]|) =
    let val pterm = print_term f
    in
	fn (ATOM (P, l)) =>
	   let val delim = ref ""
	   in
	       put P;
	       put "(";
	       iterate
		 (put (!delim);
		  delim := ",";
		  pterm t)
	       | t in list l
	       end;
	       put ")"
	   end
    end;
	
fun print_definite_clause (f as |[put, ...]|) =
    let val patom = print_atom f
    in
	fn DCLAUSE (name, env, h, body) =>
	   (put "(";
	    put name;
	    put ") ";
	    patom h;
	    let val delim = ref " :- "
	    in
		iterate
		  (put (!delim);
		   delim := ",";
		   patom a)
		| a in list body
		end;
		if empty env
		    then ()
		else (delim := " [";
		      iterate
			(put (!delim);
			 delim := ",";
			 put x;
			 put ":";
			 put srt)
		      | x => srt in map env
		      end;
		      put "]");
		put ".\n"
	    end
	    )
    end;

fun print_program f =
    let val pdcl = print_definite_clause f
    in
	fn (PROG (_, pr)) =>
	   (iterate
	      iterate
		pdcl dcl
	      | dcl in list dcls
	      end
	    | P => dcls in map pr
	    end;
	      #flush f ())
    end;

fun add_gclause (pr, dcl as DCLAUSE (_, _, ATOM (P, _), _)) =
    if P inset pr
	then pr ++ {P => dcl :: ?pr P}
    else pr ++ {P => [dcl]};

fun atom_subst sigma =
    let val ts = tsubst sigma
    in
	fn ATOM (P, l) => ATOM (P, [ts t | t in list l])
    end;

fun definite_clause_subst sigma =
    let val asb = atom_subst sigma
    in
	fn (env', DCLAUSE (name, env, h, body)) =>
	   DCLAUSE (name, (sigma <-| env) ++ env',
		    asb h,
		    [asb a | a in list body])
    end;
