
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "evaterm_h"

exception NotColor; (* raised when a color was expected (see below, comment on find),
		     but a term was provided. *)
exception NotRewrite of string; (* raised when add_rewrite is given a rule u->v
				 with some variable x free in v but not in u
				 (raised with x as argument). *)

type eva_color = eva_term;

type 'a cc = |[ find : eva_term -> eva_color, (* yields a unique representative of each term (its _color_). *)
		equate : eva_term * eva_term -> unit, (* equate two terms *)
		class : eva_color -> eva_term set, (* return equivalence class denoted by color. *)
		select : string set * eva_color -> eva_term set, (* return set of terms T_APP(f,_) in
								  equivalence class with f in given set. *)
		add_rewrite : eva_term * eva_term * (string -m> 'a) -> unit
		(* add_rewrite (u, v, vars) instantiates the rewrite rule forall x1..xn . u = v,
		 where vars={x1,...,xn}, to all instances u.sigma = v.sigma such that u.sigma is
		 in the current universe, and equates u.sigma with v.sigma.  This may expand the
		 universe, if v.sigma is not already in it. *)
		]|;

extern val cc : unit -> 'a cc;
